//////////////////////////////////////////////////////////////////////////////
// Sketch: usb_ghost
// แสดงตัวอย่างการจำลองบอร์ดไมโครคอนโทรลเลอร์ให้เป็นอุปกรณ์คีย์บอร์ด เมาส์ และเกมแพด
// ในเวลาเดียวกัน
//
// == การคอมไพล์ ==
// ต้องการไลบรารี pt และ usbdrv (มีมาให้กับแพคเกจ cpeku-practicum เรียบร้อยแล้ว)
//
// == การใช้งาน ==
// ทำแสงบริเวณตัวรับแสงให้มืดเพื่อสั่งให้บอร์ดเริ่มส่งคีย์และขยับเมาส์
//
// == หมายเหตุ ==
// อุปกรณ์ USB ที่จำลองขึ้นกำหนดให้มีหมายเลข VID:PID เป็น 16c0:0482 ซึ่งได้รับการจดทะเบียน
// ให้ใช้งานเป็น Keyboard/Mouse/Joystick combo โดย VOTI
//////////////////////////////////////////////////////////////////////////////
#include <pt.h>
#include <usbdrv.h>
#include "keycodes.h"

#define REPORT_ID_KEYBOARD 1
#define RQ_SET_LED    0

// มาโครสำหรับจำลองการหน่วงเวลาใน protothread
#define PT_DELAY(pt,ms,tsVar) \
  tsVar = millis(); \
  PT_WAIT_UNTIL(pt, millis()-tsVar >= (ms));

#define ROWS 3 
#define COLS 3

PROGMEM const char usbHidReportDescriptor[USB_CFG_HID_REPORT_DESCRIPTOR_LENGTH] = 
{
  ///////////////////
  // Keyboard
  ///////////////////
  0x05, 0x01,                    // USAGE_PAGE (Generic Desktop)
  0x09, 0x06,                    // USAGE (Keyboard)
  0xa1, 0x01,                    // COLLECTION (Application)
  0x85, 0x01,                    //   REPORT_ID (1)
  0x05, 0x07,                    //   USAGE_PAGE (Keyboard)
  0x19, 0xe0,                    //   USAGE_MINIMUM (Keyboard LeftControl)
  0x29, 0xe7,                    //   USAGE_MAXIMUM (Keyboard Right GUI)
  0x15, 0x00,                    //   LOGICAL_MINIMUM (0)
  0x25, 0x01,                    //   LOGICAL_MAXIMUM (1)
  0x75, 0x01,                    //   REPORT_SIZE (1)
  0x95, 0x08,                    //   REPORT_COUNT (8)
  0x81, 0x02,                    //   INPUT (Data,Var,Abs)
  0x95, 0x01,                    //   REPORT_COUNT (1)
  0x75, 0x08,                    //   REPORT_SIZE (8)
  0x25, 0x65,                    //   LOGICAL_MAXIMUM (101)
  0x19, 0x00,                    //   USAGE_MINIMUM (Reserved (no event indicated))
  0x29, 0x65,                    //   USAGE_MAXIMUM (Keyboard Application)
  0x81, 0x00,                    //   INPUT (Data,Ary,Abs)
  0xc0,                          // END_COLLECTION
};

struct ReportKeyboard
{
  /* +----\------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |Byte \ Bit |  7  |  6  |  5  |  4  |  3  |  2  |  1  |  0  |
   * +------\----+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  0        |               Report ID = 1                   |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  1        |           Modifiers (shift,ctrl,etc)          |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   * |  2        |                 Key Code                      |
   * +-----------+-----+-----+-----+-----+-----+-----+-----+-----+
   */
  uint8_t  report_id;
  uint8_t  modifiers;
  uint8_t  key_code;
};

ReportKeyboard reportKeyboard;

bool keys[ROWS][COLS] = {
  {1,1,1},
  {1,1,1},
  {1,1,1},
};

byte rowSW[ROWS] = { PIN_PC5, PIN_PC3, PIN_PC1 };
byte colSW[COLS] = { PIN_PB5, PIN_PB3, PIN_PB1 }; 

byte rowLED[ROWS] = {PIN_PC0, PIN_PC2, PIN_PC4};
byte colLED[COLS] = {PIN_PB4, PIN_PB2, PIN_PB0};

byte LEDs[ROWS][COLS] = {
  {1, 1, 1},
  {1, 1, 1},
  {1, 1, 1}
};

// Protothread states
struct pt main_pt;
struct pt keyboard_pt;
struct pt pt_taskSwitch;
struct pt pt_taskScreen;
struct pt pt_taskReadLED;
////////////////////////////////////////////////////////////////
// Automatically called by usbpoll() when host makes a request
////////////////////////////////////////////////////////////////
usbMsgLen_t usbFunctionSetup(uint8_t data[8])
{
  usbRequest_t *rq = (usbRequest_t*)data;
  if (rq->bRequest == RQ_SET_LED)
  {
    uint16_t led_val = rq->wValue.bytes[0] | ((rq->wValue.bytes[1])<<8);
    uint8_t led_no  = rq->wIndex.bytes[0];
    Serial.println(led_val);
    for(int i=0;i<9;i++) {
      LEDs[2-i/3][2-i%3] = (led_val&1)^1;
      led_val = led_val >> 1;
    }

    return 0;  // return no data back to host
  }

  return 0;   /* nothing to do; return no data back to host */
}

//////////////////////////////////////////////////////////////
void sendKey(uint8_t keycode, uint8_t modifiers)
{
  reportKeyboard.key_code = keycode;
  reportKeyboard.modifiers = modifiers;
  usbSetInterrupt((uchar*)&reportKeyboard, sizeof(reportKeyboard));
}

//////////////////////////////////////////////////////////////
PT_THREAD(keyboard_task(struct pt *pt))
{
  static uint32_t ts = 0;
  static uint8_t i;
  static uint8_t sequence[] = {
    KEY_T, KEY_Y, KEY_U, KEY_G, KEY_H, KEY_J, KEY_B, KEY_N, KEY_M  //Player 2
   // KEY_Q, KEY_W, KEY_E, KEY_A, KEY_S, KEY_D, KEY_Z, KEY_X, KEY_C                     //Player 1
  };

  PT_BEGIN(pt);

  while (1)
  {
    for (i = 0; i < sizeof(sequence); i++)
    {
      for(int r=0;r<ROWS;r++)
        for(int c=0;c<COLS;c++)
          if(keys[r][c] == 0){
            sendKey(sequence[r*3+c], 0);      
          }   
      PT_DELAY(pt,25,ts);
      sendKey(KEY_NONE, 0);
      // Release key
      //sendKey(KEY_NONE, 0);
      //PT_DELAY(pt,250,ts);
    }
  }
  
  PT_END(pt);
}

//////////////////////////////////////////////////////////////
PT_THREAD(taskSwitch(struct pt* pt))
{
  static uint32_t ts;

  PT_BEGIN(pt);

  while (1)
  {
    for (int r=0; r<ROWS; r++) {
      digitalWrite(rowSW[r],LOW);
      for (int c=0; c<COLS; c++){
        if(digitalRead(colSW[c]) == LOW){
          keys[r][c] = 0;
        }else{
          keys[r][c] = 1;
        }
      }
      digitalWrite(rowSW[r],HIGH);
    }
    PT_YIELD(pt);
  }

  PT_END(pt);
}

//////////////////////////////////////////////////////////////
PT_THREAD(taskScreen(struct pt* pt))
{
  static uint32_t ts;
  static int r = 0;

  PT_BEGIN(pt);

  while (1)
  {
    for (r=0;r<ROWS;r++){
      digitalWrite(rowLED[r], HIGH);
      delay(1);
      for(int c=0;c<COLS;c++){
       if(LEDs[r][c] == 0)
         digitalWrite(colLED[c], LOW);
       else
         digitalWrite(colLED[c], HIGH);
      }
      digitalWrite(rowLED[r], LOW);
    }
    PT_YIELD(pt);
  }
  PT_END(pt);
}

PT_THREAD(taskReadLED(struct pt* pt))
{
  static uint32_t ts;
 
  PT_BEGIN(pt);

  while (1)
  {
    for(int i = 0 ; i<ROWS ; i++){
      Serial.print(LEDs[i][0]);
      Serial.print(" ");
      Serial.print(LEDs[i][1]);
      Serial.print(" ");
      Serial.print(LEDs[i][2]);
      Serial.println(" ");
    }
    Serial.println("===================");
    PT_DELAY(pt, 1000, ts);
  }
 
  PT_END(pt);
}

//////////////////////////////////////////////////////////////
PT_THREAD(main_task(struct pt *pt))
{
  PT_BEGIN(pt);

  PT_WAIT_UNTIL(pt,usbInterruptIsReady());
  keyboard_task(&keyboard_pt);
  taskScreen(&pt_taskScreen);
  taskSwitch(&pt_taskSwitch);
  taskReadLED(&pt_taskReadLED);
  PT_END(pt);
}

//////////////////////////////////////////////////////////////
void setup()
{
  // Initialize USB subsystem
  usbInit();
  usbDeviceDisconnect();
  delay(300);
  usbDeviceConnect();
  
  // Initialize hardware pins
  for(int i=0;i<3;i++) {
    pinMode(colLED[i], OUTPUT);
    pinMode(rowLED[i], OUTPUT);
    digitalWrite(colLED[i], HIGH);
    digitalWrite(rowLED[i], LOW);
    pinMode(colSW[i],INPUT_PULLUP);
    pinMode(rowSW[i],OUTPUT);
    digitalWrite(rowSW[i], HIGH);
  }
  Serial.begin(9600);

  // Initialize USB reports
  reportKeyboard.report_id = REPORT_ID_KEYBOARD;
  reportKeyboard.modifiers = 0;
  reportKeyboard.key_code = KEY_NONE;

  // Initialize tasks
  PT_INIT(&main_pt);
  PT_INIT(&keyboard_pt);
  PT_INIT(&pt_taskSwitch);
  PT_INIT(&pt_taskScreen);
}

//////////////////////////////////////////////////////////////
void loop()
{
  usbPoll();
  main_task(&main_pt);
}
